import typing as t

import pandas as pd

from pysqltools.src.log import PabLog
from pysqltools.src.sql.query import Query

lg = PabLog(__name__)


def __get_data_points(df: pd.DataFrame) -> str:
    """
    Private function to get the dataframe fields as a SQL string
    """
    cols = df.columns
    data_points_str = ""
    for _, row in df.iterrows():
        conditions = [
            Query("{{ninja_column}} = {{ninja_value}}")
            .format(
                ninja_column=col,
                ninja_value=row[col],
            )
            .sql
            for col in cols
        ]
        statement = "OR ( " + " AND ".join(conditions) + ") "
        data_points_str += statement
    data_points_str = data_points_str[2:]
    return data_points_str


def delete_from_dataframe(
    df: pd.DataFrame,
    table: str,
    batch_size: t.Optional[int] = None,
    subset: t.Optional[list[str]] = None,
) -> t.Union[Query, t.Generator[Query, t.Any, t.Any]]:
    """
    ## delete_from_dataframe
    -------
    ### Parameters:
    -------

    - df (pd.DataFrame): The dataframe that will be used to set the condition of which table rows will be deleted
    - table (str): The target table (and schema if needed) to delete rows from
    - batch_size [Optional] (int): If a batch size is added, the function will return a generator iterating batch of
    the specified size with delete queries
    - subset [Optional] (list[str]): Subset of columns of the dataframe to use on the delete conditions

    -------
    ### Output:
    -------
    If the batch_size is not specified, the function will return a Query object with an SQL Ready to delete all the rows
    on the table that currently are on the dataframe

    If the batch_size is specified, the function will return a Generator that will yield Query objects containing the deletion of
    {batch_size} rows

    """
    if not df.empty:
        if batch_size is None:
            return Query(f"""DELETE FROM {table} WHERE {data_points_str}""")
        previous_iteration = 0
        while previous_iteration < len(df):
            df_batch = df.iloc[previous_iteration : previous_iteration + batch_size]
            data_points_str = __get_data_points(df_batch)
            yield Query(f"""DELETE FROM {table} WHERE {data_points_str}""")



"""Logger"""

import logging
from typing import Any, Callable

import pandas as pd
from rich.console import Console
from rich.logging import RichHandler
from rich.markdown import Markdown
from rich.progress import Progress
from rich.table import Table

COLORS = [
    "purple",
    "magenta",
    "turquoise2",
    "blue_violet",
    "orange3",
    "dark_red",
    "yellow4",
    "green",
    "deep_pink3",
    "dark_goldenrod",
    "aquamarine1",
    "chartreuse1",
]


class PabLog:
    """Logger"""

    def __init__(
        self, log_name: str = "Logger", __format: str = "- %(message)s"
    ) -> None:
        """Logger Constructor"""
        self.avaliable_colors = []
        self.rich_handler = RichHandler(rich_tracebacks=True)

        self.avaliable_colors = COLORS

        __handlers = [self.rich_handler]

        logging.basicConfig(format=__format, handlers=__handlers, level=logging.DEBUG)

        self.log = logging.getLogger(log_name)
        self.log.setLevel(logging.INFO)
        self.console = Console()

    def add_table(self, df: pd.DataFrame, title: str = "", max_rows: int = 10) -> None:
        """Add a table in terminal"""
        table = Table(title=title)
        color = 0
        for col in df.columns:
            if color >= len(self.avaliable_colors) - 1:
                color = 0
            else:
                color += 1

            table.add_column(str(col), style=self.avaliable_colors[color])

        for _, row in df.head(max_rows).iterrows():
            table.add_row(*[str(r) for r in row])

        self.console.print(table)

    def add_md(self, markdown: str):
        """Add markdown in terminal"""
        md = Markdown(markdown)
        self.console.print(md)

    def add_title(self, title: str):
        """Add markdown title in terminal"""
        md_ = f" # **{title}** "
        md = Markdown(md_)
        self.console.print(md)


def progress_function(
    task_name: str = "Task in progress...", color: str = "green", total: int = 100
):
    """
    Decorator to use a rich progress bar
    """

    def decorator(fun: Callable[..., Any]):
        """Decorator"""

        def inner(*args, **kwargs):
            """Wrapped"""
            with Progress() as progress:
                task = progress.add_task(f"[{color}] {task_name}...", total=total)
                prog = {"progress": progress, "task": task}
                result = fun(*args, **kwargs, **prog)
                progress.update(task, advance=total)
            return result

        return inner

    return decorator



"""
Source code for the pysqltools package
"""

from pysqltools.src.sql.insert import generate_insert_query, insert_pandas
from pysqltools.src.sql.query import Query, SQLString, get_queries_from_path
from pysqltools.src.sql.table import Table



"""
This module contains all the classes and 
functions to interact with SQL text objects.
"""

import datetime
import os
import re
from typing import Any, Generator, Union

import rich
import rich.progress
import sqlparse
from multimethod import multimethod

from pysqltools.src.log import PabLog, progress_function
from pysqltools.src.sql.exceptions import QueryFormattingError

lg = PabLog("Query")


class QueryException(Exception):
    """
    Exception during Query processing.
    """

    def __init__(self, *args: object) -> None:
        super().__init__(*args)


class SQLString(str):
    """
    String Class used to format queries without adding single quotes.
    """


@multimethod
def assign_parameter(param: str) -> str:
    """
    assign parameter function
    """
    return "'" + param + "'"


@multimethod
def assign_parameter(param: Union[int, float]) -> str:
    """
    assign parameter function
    """
    return param


@multimethod
def assign_parameter(param: SQLString) -> str:
    """
    assign parameter function
    """
    return param


@multimethod
def assign_parameter(param: list[str]) -> str:
    """
    assign parameter function
    """
    string = "("
    for item in param:
        string += f"'{item}', "
    return string[:-2] + ")"


@multimethod
def assign_parameter(param: list[int, float]) -> str:
    """
    assign parameter function
    """
    string = "("
    for item in param:
        string += f"{item}, "
    return string[:-2] + ")"


@multimethod
def assign_parameter(param: datetime.datetime) -> str:
    """
    assign parameter function
    """
    return "datetime '" + param.strftime("%Y-%m-%d %H:%M:%S") + "'"


class Query:
    """
    ### Query Module
    The query module provides a Query class to work with Query objects, which will allow to modify the
    SQL Queries on an easy way with the class methods, and easily access the sql string with the sql
    attribute of the objects.
    --------------------------
    ### Parameters
    - sql: property. The string containing the SQL Query
    - parsed: sqlparse object from the SQL Query
    - options: via kwargs. Current options:
        - indent_query (bool) default True: Re-indent the query for output

    To add parameters to the query, use {{parameter}} on the SQL String.

    The current methods are:

    - ctes: Generator that yields the CTEs of the Query
    - selects: Generator that yields the Select statements of the Query
    - Windows: Generator that yields the Window Function contents of the query
    - tables: Generator that yields the detected tables on the query
    - parameters: Generator that yields all the parameters on the Query
    - format: allows to assign values to the parameters in the query. Current supported types are str, int, float, datetime.datetime, list[int, float, str]
    To call the format function, just call the parameters you have defined on your query. Example:
    query:
    `select * from {{table_param}} limit 20`

    function call:
    `query = Query(sql = sql).format(table_param = "MyTable")`
    """

    def __init__(self, sql: str, **kwargs) -> None:
        self.lower_query = False
        if "lower_query" in kwargs.keys():
            if kwargs.get("lower_query") == True:
                self.lower_query = True
                self._sql = sql.lower()
        else:
            self._sql = sql
        self.parsed = sqlparse.parse(sql)[0]
        self.options = kwargs
        self._parameters = None
        self._tables = None
        self._selects = None
        self._windows = None
        self._ctes = None

    @property
    def sql(self):
        """
        Contains the string with the SQL Statement. If the flag `indent_query` has been set on the
        constructor, the sql will be returned with automatic indentation.
        """
        if self.options.get("indent_query", True):
            self._sql = str(
                sqlparse.format(
                    self._sql,
                    keyword_case="lower",
                    id_case="lower",
                    indent_columns=True,
                    reindent=True,
                )
            )
        return self._sql

    @sql.setter
    def sql(self, sql: str):
        if self.lower_query:
            self._sql = sql.lower()
        else:
            self._sql = sql

    @property
    def ctes(self) -> Generator:
        """
        returns a generator containing all the CTEs on the query. The generator returns
        the CTE identifier as first argument and the CTE Content as second argument.
        """
        cte_regex = re.compile(
            r"""(?i)\b(\w+)\s+as\s+\((.*?)\)(?=\s*,|\s*select|\s*insert|\s*update|\s*delete|\s*with|\Z)""",
            re.DOTALL | re.IGNORECASE | re.MULTILINE,
        )

        matches = cte_regex.findall(self.sql)
        self._ctes = []
        for _, match in enumerate(matches, 1):
            cte_name, cte_content = match
            self._ctes.append((cte_name, cte_content))

        yield from self._ctes

    @ctes.setter
    def ctes(self) -> None:
        raise QueryException("ctes is read-only")

    @property
    def parameters(self) -> Generator:
        """returns a generator containing all the Parameters on the query.
        Parameters must be between {{ }}"""
        regex = re.compile(
            r"(?<={{)\S*(?=}})",
            re.DOTALL | re.IGNORECASE | re.MULTILINE,
        )
        self._parameters = regex.findall(self.sql)
        yield from self._parameters

    @parameters.setter
    def parameters(self) -> None:
        raise QueryException("parameters is read-only")

    @property
    def tables(self) -> Generator:
        """Returns a generator containing all the detected tables. CTEs identifiers excluded"""
        regex = re.compile(
            r"(?<=from|join).*?\s*\S*",
            re.DOTALL | re.IGNORECASE | re.MULTILINE,
        )
        self._tables = [r.strip() for r in regex.findall(self.sql)]
        self._tables = [
            t for t in self._tables if t not in [c[0] for c in list(self.ctes)]
        ]
        yield from self._tables

    @tables.setter
    def tables(self, *args: Any) -> None:
        raise QueryException("tables is read-only")

    @property
    def selects(self) -> Generator:
        """returns a generator containing all the Select contents on the query"""
        self._selects = [i.strip() for i in self.__non_greedy_regex("select", "from")]
        yield from self._selects

    @selects.setter
    def selects(self) -> None:
        raise QueryException("selects is read-only")

    @property
    def windows(self) -> Generator:
        """returns a generator containing all the Window Functions on the query"""
        self._windows = [i.strip() for i in self.__non_greedy_regex("over", r"\)")]
        yield from self._windows

    @windows.setter
    def windows(self) -> None:
        raise QueryException("windows is read-only")

    def __non_greedy_regex(self, keyword_start: str, keyword_end: str) -> Generator:
        """"""
        regex = re.compile(
            rf"(?<={keyword_start}).*?(?={keyword_end})",
            re.DOTALL | re.IGNORECASE | re.MULTILINE,
        )
        yield from [i.strip() for i in regex.findall(self.sql)]

    def iter_query_lines(self) -> Generator:
        """
        Iterate the SQL Query line by line.
        """
        yield from self.sql.split("\n")

    def get_ctes_dict(self) -> Generator:
        """
        Get the CTEs on a dictionary

        """
        cte_dict = {}
        for k, v in self.ctes:
            cte_dict.update({k: v})
        return cte_dict

    def format(self, **kwargs) -> "Query":
        """
        Allows dynamic variables on SQL Queries.
        The parameters must be between keys i.e. {{parameter}}. Using the format function,
        you can substitute the parameters with python variables.\n
        An special type SQLString is used for tables, as we don't want to include "'" on
        those strings.\n

        `sql = Query(sql="select * from {{my_table}} where country in {{country_list}}")` \n
        `sql.format(my_table=SQLString("schema.table"), country_list = ['ES', 'GB'])`
        """
        try:
            for k, v in kwargs.items():
                self.sql = self.sql.replace("{{" + k + "}}", assign_parameter(v))
            self.parsed = sqlparse.parse(sql=self.sql)[0]
        except Exception as e:
            raise QueryFormattingError(e)

        return self

    def get_cte_by_identifier(self, identifier: str) -> Union[None, str]:
        """
        Pass the identifier of one of the query CTEs and get the string containing the content of the CTE.
        """
        if identifier in self.get_ctes_dict():
            return self.get_ctes_dict().get(identifier)
        else:
            return None

    def replace_cte(self, identifier: str, new_cte_content: str) -> "Query":
        """
        Given a CTE identifier, change its content with a new string
        """
        if identifier in self.get_ctes_dict():
            self.sql = self.sql.replace(
                self.get_cte_by_identifier(identifier), new_cte_content
            )
            return self
        else:
            raise TypeError(f"CTE with dentifier {identifier} not found in Query")

    def __str__(self):
        return self.sql

    def __dict__(self):
        return {
            "tables": list(self.tables),
            "ctes": self.get_ctes_dict(),
            "parameters": list(self.parameters),
        }


@progress_function("searching queries...", color="green")
def get_queries_from_path(path: str = None, *args, **kwargs) -> list[Query]:
    # lg.add_md("## Scanning Directory for SQL Queries")
    queries = {}
    for dirpath, dirname, filename in os.walk(path):
        lg.log.info(f"Searching {dirpath}...")
        kwargs["progress"].advance(kwargs["task"], 1)
        i = 2
        for f in filename:
            if f.__contains__(".sql"):
                lg.log.info(f"Added item '{f}'")
                if f in queries:
                    name = f.replace(".sql", "") + f"_{i}"
                    i += 1
                else:
                    name = f.replace(".sql", "")
                queries.update(
                    {
                        name: Query(
                            open(os.path.join(dirpath, f), "r", encoding="utf-8").read()
                        )
                    }
                )

    return queries



"""
This module allows to interact with tables in SQL (Create, Alter, etc)
"""

from typing import Any, Union

import pandas as pd
import sqlparse

from pysqltools.src.sql.query import Query

from .constants import TYPE_MAPPING
from .insert import insert_pandas


class Table:
    """
    Class to manipulate SQL Tables
    """

    def __init__(self, table: str, schema: Union[str, None] = None) -> None:
        self.table = table
        if schema:
            self.table = f"{schema}.{table}"

    def create_from_df(
        self,
        df: pd.DataFrame,
        execute: bool = False,
        insert_data: bool = False,
        **insert_kwargs: Any,
    ) -> str:
        """
        Get the SQL statement to create a SQL table based on a Pandas DataFrame. If the insert_data argument is set to True,
        you must pass the same arguments as to the `pysqltools.SQL.insert_pandas` as **kwargs. Example:
        ```python
        table = Table(table = "myTable", schema = "dbo")
        table.create_from_df(df, insert_data = True, connection = myConnection, batch_size = 10000)
        ```
        """
        columns = dict(
            zip(
                df.dtypes.index.to_list(),
                map(
                    lambda x: TYPE_MAPPING[x],
                    map(lambda x: str(x), df.dtypes.to_list()),
                ),
            )
        )
        sql = f"CREATE TABLE IF NOT EXISTS {self.table} ( "
        for k, v in columns.items():
            sql += f"{k} {v}, "
        sql = sql[:-2] + " )"
        if not insert_data:
            return sqlparse.format(sql, encoding="utf-8")
        if execute:
            insert_kwargs["connection"].execute(Query(sql))
        if "batch_size" in insert_kwargs:
            batch_size = insert_kwargs["batch_size"]
        else:
            batch_size = 1000
        try:
            insert_pandas(
                df,
                connection=insert_kwargs["connection"],
                table=self.table,
                batch_size=batch_size,
                dialect=insert_kwargs["dialect"],
            )
        except TypeError as e:
            raise TypeError(
                "Please include the insert arguments into the create_table_from_df method",
                e,
            )



from .src.sql.delete import delete_from_dataframe
from .src.sql.insert import generate_insert_query, insert_pandas, insert_pandas_threads
from .src.sql.query import Query, SQLString, get_queries_from_path
from .src.sql.table import Table


def format_sql(sql: str, **kwargs) -> str:
    query = Query(sql=sql).format(kwargs)
    return query.sql



from datetime import date
from multiprocessing import Process, Queue
from typing import Any, Callable, Generator

import pandas as pd
import sqlparse
from rich.progress import Progress

from pysqltools.src.log import PabLog
from pysqltools.src.sql.query import Query

lg = PabLog("Insert")


def prepare_value(val: Any, dialect: str) -> Any:
    """
    Format value from Python types to SQL Types
    """
    if dialect.lower().__contains__("trino"):
        if isinstance(val, str):
            return f"'{val}'"
        if isinstance(val, bool):
            return val
        if isinstance(val, dict):
            val = str(val).replace("'", '"')
        if isinstance(val, pd.Timestamp):
            val = "TIMESTAMP '" + str(val) + "'"
        if isinstance(val, date):
            val = "DATE '" + str(val) + "'"
        if isinstance(val, list):
            val = "ARRAY " + str(val)
        if isinstance(val, float):
            val = "DOUBLE '" + str(val) + "'"
        if isinstance(val, int):
            val = "INT '" + str(val) + "'"
        if pd.isnull(val):
            val = "NULL"

        try:
            if (
                "'" in val
                and "DOUBLE" not in val
                and "INT" not in val
                and "TIMESTAMP" not in val
                and "DATE" not in val
            ):
                val = val.replace("'", "''")
        except TypeError:
            lg.log.warning("Not Adding Quotes")

        return val
    if dialect.lower().__contains__("mysql"):
        if isinstance(val, str):
            return f"'{val}'"
        if isinstance(val, bool):
            return bool(val)
        if isinstance(val, dict):
            val = str(val).replace("'", '"')
        if isinstance(val, pd.Timestamp):
            val = f"'{val}'"
        if isinstance(val, date):
            val = f"'{val}'"
        if isinstance(val, list):
            val = f"'{str(val)}'"
        if isinstance(val, float):
            val = str(val)
        if isinstance(val, int):
            val = str(val)
        if pd.isnull(val):
            val = "NULL"

        return val

    if dialect.lower().__contains__("sqlite"):
        if isinstance(val, str):
            return f"'{val}'"
        if isinstance(val, bool):
            return bool(val)
        if isinstance(val, dict):
            val = str(val).replace("'", '"')
        if isinstance(val, pd.Timestamp):
            val = f"'{val}'"
        if isinstance(val, date):
            val = f"'{val}'"
        if isinstance(val, list):
            val = f"'{str(val)}'"
        if isinstance(val, float):
            val = str(val)
        if isinstance(val, int):
            val = str(val)
        if pd.isnull(val):
            val = "NULL"

        return val

    if dialect.lower().__contains__("ibm"):
        if isinstance(val, str):
            return f"'{val}'"
        if isinstance(val, bool):
            return bool(val)
        if isinstance(val, dict):
            val = str(val).replace("'", '"')
        if isinstance(val, pd.Timestamp):
            val = f"TIMESTAMP '{val}'"
        if isinstance(val, date):
            val = f"DATE '{val}'"
        if isinstance(val, list):
            val = f"'{str(val)}'"
        if isinstance(val, float):
            val = str(val)
        if isinstance(val, int):
            val = str(val)
        if pd.isnull(val):
            val = "NULL"

        return val

    if dialect.lower().__contains__("sqlserver"):
        if isinstance(val, str):
            return f"'{val}'"
        if isinstance(val, bool):
            return bool(val)
        if isinstance(val, dict):
            val = str(val).replace("'", '"')
        if isinstance(val, pd.Timestamp):
            val = f"'{val}'"
        if isinstance(val, date):
            val = f"'{val}'"
        if isinstance(val, list):
            val = f"'{str(val)}'"
        if isinstance(val, float):
            val = str(val)
        if isinstance(val, int):
            val = str(val)
        if pd.isnull(val):
            val = "NULL"

        return val

    if dialect.lower().__contains__("mariadb"):
        if isinstance(val, str):
            return f"'{val}'"
        if isinstance(val, bool):
            return bool(val)
        if isinstance(val, dict):
            val = str(val).replace("'", '"')
        if isinstance(val, pd.Timestamp):
            val = f"'{val}'"
        if isinstance(val, date):
            val = f"'{val}'"
        if isinstance(val, list):
            val = f"'{str(val)}'"
        if isinstance(val, float):
            val = str(val)
        if isinstance(val, int):
            val = str(val)
        if pd.isnull(val):
            val = "NULL"
        return val


def join_values(data: list[Any], dialect: str) -> str:
    """
    Create a String for the VALUES () SQL Syntax
    """
    clean_list = []
    for val in data:
        try:
            if "NULL" in val:
                val = "NULL"
        except TypeError:
            lg.log.error("Type not using Nulls")

        clean_list.append(str(val))
    str_data = (", ").join(clean_list)
    return "(" + str_data + ")"


def pandas_to_sql(df: pd.DataFrame, dialect: str) -> Generator[str, None, None]:
    """
    Generator to get one row insert statement
    """
    for row in df.values:
        data_list = [prepare_value(x, dialect=dialect) for x in row]
        data_string = join_values(data_list, dialect=dialect)
        yield data_string


def generate_insert_query(
    df: pd.DataFrame,
    table: str = None,
    schema: str = None,
    batch_size: int = 5000,
    dialect: str = "trino",
) -> Generator[Query, None, None]:
    """
    Generates a query to insert the data into the database

    Parameters:
    ------------------------------------------------------------------------------
    - df (pd.DataFrame): The DataFrame to be inserted
    - table (str, optional): The name of the table to insert into (default is None)
    - schema (str, optional): The schema for the table (default is None)
    - batch_size (int, optional): The number of rows to insert in each batch (default is 5000)
    - dialect (str, optional): The SQL dialect (default is "trino")
    """
    if df.empty:
        raise TypeError("DataFrame can not be empty")
    previous_iter = 0
    while previous_iter < len(df):
        percentage = round(100 * previous_iter / len(df), 2)
        lg.log.info("Generating Insert Queries... %s", percentage)
        batch = df.iloc[previous_iter : previous_iter + batch_size]
        data_points = list(pandas_to_sql(batch, dialect))
        data_points_string = ",".join(data_points)
        if schema and table:
            table = f"{schema}.{table}"
        if not table:
            table = "{{table}}"
        sql = f"INSERT INTO {table} VALUES {data_points_string}"
        query = Query(sqlparse.format(sql))
        previous_iter += batch_size
        yield query


def insert_pandas(
    df: pd.DataFrame,
    batch_size: int,
    table: str,
    execute_function: Callable[..., Any],
    schema: str = "",
    dialect: str = "trino",
) -> None:
    """
    Insert a pandas DataFrame into a specified table using a generator of queries

    Parameters:
    ------------------------------------------------------------------------------
    - df (pd.DataFrame): The DataFrame to be inserted
    - batch_size (int): The number of rows to insert in each batch
    - table (str): The name of the table to insert into
    - execute_function (Callable[..., Any]): The function to execute the query
    - schema (str, optional): The schema for the table (default is "")
    - dialect (str, optional): The SQL dialect (default is "trino")
    """
    if not table and schema:
        raise TypeError("Table and Schema need to be provided")
    with Progress() as progress:
        iterations = len(df) / batch_size
        task1 = progress.add_task("[red] Generating Queries...", total=1000)
        task2 = progress.add_task("[green] Inserting Data...", total=iterations)
        task3 = progress.add_task("[cyan] Finishing...", total=1000)
        for _ in range(1000):
            progress.update(task1, advance=1.0)
        for query in generate_insert_query(
            df, table, schema, batch_size, dialect=dialect
        ):
            try:
                execute_function(query)
            except Exception as e:
                lg.log.warning("Query Execution Failed")
                lg.log.error(e)
                lg.log(query.sql)
            progress.update(task2, advance=1)
        for i in range(1000):
            progress.update(task3, advance=1.0)


def insert_pandas_threads(
    df: pd.DataFrame,
    batch_size: int,
    table: str,
    execute_function: Callable[..., Any],
    schema: str = "",
    dialect: str = "trino",
    threads: int = 2,
) -> None:
    """
    Insert a pandas DataFrame into a specified table using a generator of queries,
    and using multiprocessing to accelerate the process.

    Parameters:
    ------------------------------------------------------------------------------
    - df (pd.DataFrame): The DataFrame to be inserted
    - batch_size (int): The number of rows to insert in each batch
    - table (str): The name of the table to insert into
    - execute_function (Callable[..., Any]): The function to execute the query
    - schema (str, optional): The schema for the table (default is "")
    - dialect (str, optional): The SQL dialect (default is "trino")
    """

    def process_queue(q: Queue) -> None:
        item = q.get()
        execute_function(item)

    queries = list(
        generate_insert_query(df, table, schema, batch_size, dialect=dialect)
    )

    queue = Queue()
    process = Process(target=process_queue, args=(queue,))
    process.start()
    process.join()



import datetime

import yaml


class Dependency:
    """
    A Dependency represents a dependency between two data sources.
    Attributes:
        source (str): The name of the source data source.
        target (str): The name of the target data source.
        freshness (datetime.datetime): The freshness timestamp of the dependency.
        dependency_type (str): The type of dependency (e.g., "upstream", "downstream").
        description (str): A brief description of the dependency.
    """


class Sensor:
    """
    A Sensor represents a data source with a table and a freshness timestamp.
    Attributes:
        table (str): The name of the table in the database.
        freshness (datetime.datetime): The freshness timestamp of the data source.
    Methods:
        update_freshness(): Updates the freshness timestamp to the current time.
        to_yaml(): Returns a YAML representation of the Sensor.

    """

    def __init__(self, table: str) -> None:
        self._freshness: datetime.datetime
        self.table = table

    @staticmethod
    def from_yaml(yaml_file_path: str) -> "Sensor":
        try:
            with open(yaml_file_path, "r") as file:
                data = yaml.safe_load(file)
            sensor = Sensor(table=data["table"])
            sensor._freshness = datetime.datetime.fromisoformat(data["freshness"])
            return sensor
        except (yaml.YAMLError, KeyError, ValueError) as e:
            raise Exception(f"Invalid YAML content: {e}")

    @property
    def freshness(self) -> datetime.datetime:
        return self._freshness

    @freshness.setter
    def freshness(self, value: datetime.datetime) -> None:
        raise Exception("Freshness can not be manually set")

    def update_freshness(self) -> "Sensor":
        self._freshness = datetime.datetime.now()
        return self

    def to_yaml(self):
        return yaml.dump(
            {
                "table": self.table,
                "freshness": self._freshness.isoformat(),
            }
        )



"""
Constant values for the queries package
"""

KEYWORDS = ["select", "from", "where", "group", "having", "order", "limit"]

TYPE_MAPPING = {
    "object": "varchar",
    "int64": "int",
    "float64": "double",
    "bool": "bool",
    "datetime64": "timestamp",
    "datetime64[ns]": "timestamp",
}



"""
Queries Package. Contains everything SQL-Text query related
"""

from pysqltools.src.sql.delete import delete_from_dataframe
from pysqltools.src.sql.insert import (
    generate_insert_query,
    insert_pandas,
    insert_pandas_threads,
)
from pysqltools.src.sql.query import Query, SQLString, get_queries_from_path
from pysqltools.src.sql.table import Table
